/*
 * nsVanadielTimer.js
 *
 * Copyright (c) 2005, 2006 by Daniel Potter.  All rights reserved.
 *
 * This file is part of the Vana'diel Clock.
 *
 * The Vana'diel Clock is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The Vana'diel Clock is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Vana'diel Clock; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Creates a timer service that ticks every Vana'diel minute, updating any
 * callback functions with new time information.
 */

const DAYS    = 24 * 60 * 60;
const HOURS   = 60 * 60;
const MINUTES = 60;

function VanadielTimer() {
    this.mListeners = new Array();

    this.mPrefService = Components.classes["@mozilla.org/preferences-service;1"].
	    getService(Components.interfaces.nsIPrefBranch);
    this.reset();
}

VanadielTimer.prototype = {
    mIsActive: false,
    mOffset: 0,
    mRealTimeOffset: 0,
    mTimer: null,
    mObserverService: null,
    mPrefService: null,
    mLastDay: 0,
    mVanaYear: 0,
    mVanaMonth: 0,
    mVanaDay: 0,
    mVanaDayWeek: 0,
    mVanaHour: 0,
    mVanaMin: 0,
    mVanaTime: 0,
    mBasicMoonPhase: 0,
    mMoonPercent: 0,
    mMoonPhase: 0,
    mWaxing: true,
    mListeners: null,
    mAlarms: null,
    mResync: false,

    get active()          { return this.mIsActive; },
    get vanadielOffset()  { return this.mOffset; },
    get year()            { return this.mVanaYear; },
    get month()           { return this.mVanaMonth; },
    get day()             { return this.mVanaDay; },
    get weekday()         { return this.mVanaDayWeek; },
    get hour()            { return this.mVanaHour; },
    get minute()          { return this.mVanaMin; },
    get time()            { return this.mVanaTime; },
    get epochDays()       { return this.mLastDay; },
    get moonPhase()       { return this.mMoonPhase; },
    get moonPercent()     { return this.mMoonPercent; },
    get basicMoonPhase()  { return this.mBasicMoonPhase; },
    get moonPhaseWaxing() { return this.mWaxing; },
    get realTimeOffset()  { return this.mRealTimeOffset; },

    start: function () {
	if (!this.mIsActive) {
	    this.mIsActive = true;
	    this.mTimer = Components.classes["@mozilla.org/timer;1"]
			   .createInstance(Components.interfaces.nsITimer);
	    // Try and update at the next 2400ms boundary
	    this.mTimer.initWithCallback(this, 2400 - (new Date().getTime() % 2400), Components.interfaces.nsITimer.TYPE_ONE_SHOT);

	    this.mAlarms = Components.classes["@xenoveritas.org/vanadiel-alarm-manager;1"]
			       .getService(Components.interfaces.nsIVanadielAlarmManager);
	    this.mAlarms.loadAlarms();
	    // since we just started, wake everyone up
	    try {
		this.mAlarms.activateAlarms(this.mVanaTime, true);
	    } catch (ex) {
		// TODO: send JS errors to the JavaScript console?
		dump("Alarm manager caused exception:\n");
		dump(ex);
		dump("\n");
		for (var key in ex) {
		    dump("exception.");
		    dump(key);
		    dump(": ");
		    dump(ex[key]);
		    dump("\n");
		}
	    }
	}
    },

    stop: function () {
	if (this.mIsActive) {
	    this.mIsActive = false;
	    this.mTimer.cancel();
	    this.mTimer = null;
	}
    },

    reset: function () {
	var restart = this.mIsActive;
	if (restart) {
	    this.stop();
	}
	// Calculate the offset to convert to Vana'diel time for the current time zone

	// I, unfortunately, have no idea who originally came up with the following.
	// Note that it has since been converted to GMT.

	// Earth time:       2002/06/23 15:00:00 GMT
	// Vana'diel time:   0898/02/01 00:00:00 (VST?)

	var vanaTime  = (898 * 360 + 30) * 24 * 60 * 60 * 1000 / 25;
	var baseDay = new Date();
	baseDay.setUTCFullYear(2002, 5, 23); // Set date to 2002-06-23
	baseDay.setUTCHours(15, 0, 0, 0);    // Set time to 15:00:00.0000

	this.mOffset = baseDay.getTime() - vanaTime;
	// pull in the "real time offset" off the prefs
	this.mRealTimeOffset = this.mPrefService.getIntPref("vanadiel.clock.offset");

	// immediately create the current time
	this.notify(null);
	if (restart) {
	    this.start();
	}
    },

    test: function (time) {
	if (!this.mIsActive) {
	    this.setVars(time);
	}
    },

    setVars: function (vanadielTime) {
	this.mVanaTime = vanadielTime;
	// See if the day has changed
	var tDay = Math.floor(vanadielTime / DAYS);
	var updateMsg;
	if (tDay != this.mLastDay) {
	    updateMsg = "date";
	    this.mLastDay = tDay;
	    this.mVanaYear = Math.floor(tDay / 360);
	    this.mVanaMonth = Math.floor((tDay % 360) / 30) + 1;
	    this.mVanaDay = (tDay % 30) + 1;
	    this.mVanaDayWeek = tDay % 8;
	    // calculate the current moon phase - unlike the current
	    // day of the week, this isn't nice and happily rooted at 0
	    // moonphase = 0 - full moon
	    // moonphase = 42 - new moon
	    this.mMoonPhase = (tDay + 26) % 84;
	    // translate that value into a percentage value
	    this.mMoonPercent = ((42 - this.mMoonPhase) * 100) / 42;
	    // negative values are "waxing" and backwards
	    this.mWaxing = false;
	    if (this.mMoonPercent < 0) {
		this.mWaxing = true;
		this.mMoonPercent = Math.abs(this.mMoonPercent);
	    }
	    // Now make it an interger
	    this.mMoonPercent = Math.floor(this.mMoonPercent + 0.5);
	    // Translate the phase into one of the eight phases.
	    //
	    // To see how the "phase counter" stacks up to a percent, run the
	    // following Perl 1-liner:
	    //
	    // perl -e 'for($p=0;$p<84;$p++){print "$p = ".int((abs((42-$p)*100/42)+0.5))."\% (wa".($p<42?"n":"x")."ing)\n";}'
	    //
	    // Simple, huh?
	    //
	    // === MOON PHASES ===
	    // 80- 2 == 90%(Wax)  - 95%(Wane) - Full
	    //  3-16 == 93%(Wane) - 62%(Wane) - Waning Gibbous
	    // 17-23 == 60%(Wane) - 45%(Wane) - Last Quarter
	    // 24-37 == 43%(Wane) - 12%(Wane) - Waning Crescent
	    // 38-44 == 10%(Wane) -  5%(Wax)  - New Moon
	    // 45-58 ==  7%(Wax)  - 38%(Wax)  - Waxing Crescent
	    // 59-65 == 40%(Wax)  - 55%(Wax)  - First Quarter
	    // 66-79 == 57%(Wax)  - 88%(Wax)  - Waxing Gibbous
	    if (this.mMoonPhase >= 38)
		this.mBasicMoonPhase = Math.floor((this.mMoonPhase - 38) / 7);
	    else
		this.mBasicMoonPhase = Math.floor((this.mMoonPhase + 46) / 7);
	} else {
	    updateMsg = "";
	}
	this.mVanaHour = Math.floor(vanadielTime / HOURS) % 24;
	this.mVanaMin = Math.floor(vanadielTime / MINUTES) % 60;
    },

    notify: function (timer) {
	// Calculate Vana'diel time, and then fire off an event - the +500 after getTime()
	// rounds this to the nearest second.
	this.setVars((Math.floor(new Date().getTime() + this.mRealTimeOffset + 500 - this.mOffset) / 1000) * 25);
	// Notfiy observers that a new Vana'diel minute is upon us
	for (var l in this.mListeners) {
	    try {
		this.mListeners[l].vanadielTicked(this.mVanaYear, this.mVanaMonth,
			this.mVanaDay, this.mVanaDayWeek,
			this.mVanaHour, this.mVanaMin, this.mResync);
	    } catch (ex) {
		// alert on error
		// TODO: send JS errors to the JavaScript console?
		dump("Listener caused exception:\n");
		dump(ex);
		dump("\n");
		for (var key in ex) {
		    dump("exception.");
		    dump(key);
		    dump(": ");
		    dump(ex[key]);
		    dump("\n");
		}
	    }
	}
	this.mResync = false;
	// Now try and sound any alarms that have been set.
	try {
	    // don't try anything if the alarm hasn't be activated yet
	    if (this.mAlarms)
		this.mAlarms.activateAlarms(this.mVanaTime, false);
	} catch (ex) {
	    // alert on error
	    // TODO: send JS errors to the JavaScript console?
	    dump("Alarm manager caused exception:\n");
	    dump(ex);
	    dump("\n");
	    for (var key in ex) {
		dump("exception.");
		dump(key);
		dump(": ");
		dump(ex[key]);
		dump("\n");
	    }
	}
	// Now set us to fire at the next 2400ms boundary, so that we stay as close
	// to in-sync with Vana'diel time as possible
	if (this.mTimer != null) {
	    this.mTimer.cancel();
	    this.mTimer.initWithCallback(this, 2400 - (new Date().getTime() % 2400), Components.interfaces.nsITimer.TYPE_ONE_SHOT);
	}
    },

    getVanadielDate: function (year, month, day, dayWeek) {
	year.value = this.mVanaYear;
	month.value = this.mVanaMonth;
	day.value = this.mVanaDay;
	dayWeek.value = this.mVanaDayWeek;
    },

    getVanadielTime: function(hour, minute) {
	hour.value = this.hour;
	minute.value = this.minute;
    },

    addVanadielClockListener: function(listener) {
	for (var l in this.mListeners) {
	    if (this.mListeners[l] == listener)
		return;
	}
	this.mListeners.push(listener);
	//dump("Vana'diel Timer: Added listener, up to " + this.mListeners.length + " listeners\n");
    },

    removeVanadielClockListener: function(listener) {
	for (var l = 0; l < this.mListeners.length; l++) {
	    if (this.mListeners[l] == listener) {
		this.mListeners.splice(l, 1);
		//dump("Vana'diel Timer: Removed listener, down to " + this.mListeners.length + " listeners\n");
		return;
	    }
	}
	dump("Vana'diel Timer: Couldn't removed listener, listener does not exist in list\n");
    },

    /**
     * Sets the offset in milliseconds from the computer time to "real" time.
     */
    setRealTimeOffset: function(offset) {
	if (offset != null) {
	    if (isNaN(offset)) {
		return;
	    }
	    dump("Offset is now " + offset + "ms\n");
	    this.mRealTimeOffset = offset;
	    this.mPrefService.setIntPref("vanadiel.clock.offset", this.mRealTimeOffset);
	    this.mResync = true;
	    this.notify(this.mTimer);
	}
    },

    QueryInterface: function (iid) {
        if (!iid.equals(Components.interfaces.nsIVanadielTimer)
            && !iid.equals(Components.interfaces.nsISupports))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
}

/**********************************************************
   XPCOM registration stuff, from
   http://books.mozdev.org/html/mozilla-chp-8-sect-2.html
 **********************************************************/

var Module = {
    firstTime: true,

    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        compMgr =
	    compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	compMgr.registerFactoryLocation(this.myCID,
				    this.myName,
				    this.myProgID,
				    fileSpec,
				    location,
				    type);
    },

    getClassObject : function (compMgr, cid, iid) {
	if (!cid.equals(this.myCID))
	    throw Components.results.NS_ERROR_NO_INTERFACE;
        if (!iid.equals(Components.interfaces.nsIFactory))
	    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	return this.myFactory;
    },

    myCID: Components.ID("{e8513bde-bad9-4172-a67f-8de617a55a54}"),
    myProgID: "@xenoveritas.org/vanadiel-timer;1",
    myName: "Vana'diel Timer",
	
    myFactory: {
	createInstance: function (outer, iid) {
	    if (outer != null)
		throw Components.results.NS_ERROR_NO_AGGREGATION;
	    return (new VanadielTimer()).QueryInterface(iid);
	}
    },

    canUnload: function(compMgr) {
        return true;
    }
}; // END Module

function NSGetModule(compMgr, fileSpec) { return Module; }